@@ -19,6 +19,8 @@
 #include <config.h>
 #include <libgnome/libgnome.h>
 #include <signal.h>
+#include <limits.h>
+#include <stdlib.h>
 #include <unistd.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -361,6 +363,7 @@ gdm_config_parse (void)
     if (ve_string_empty (GdmServAuthDir)) {
 	    gdm_text_message_dialog
 		    (_("No daemon/ServAuthDir specified in the configuration file"));
+	    GdmPidFile = NULL;
 	    gdm_fail (_("%s: No authdir specified."), "gdm_config_parse");
     }
 
@@ -390,6 +393,8 @@ gdm_config_parse (void)
 			    (GDM_KEY_SERVER_FLEXIBLE);
 		    svr->choosable = gnome_config_get_bool
 			    (GDM_KEY_SERVER_CHOOSABLE);
+		    svr->handled = gnome_config_get_bool
+			    (GDM_KEY_SERVER_HANDLED);
 
 		    if (ve_string_empty (svr->command)) {
 			    gdm_error (_("%s: Empty server command, "
@@ -419,6 +424,7 @@ gdm_config_parse (void)
 	    svr->command = g_strdup (GdmStandardXServer);
 	    svr->flexible = TRUE;
 	    svr->choosable = TRUE;
+	    svr->handled = TRUE;
 
 	    xservers = g_slist_append (xservers, svr);
     }
@@ -499,6 +505,7 @@ gdm_config_parse (void)
 			       "and restart gdm."),
 			     GDM_CONFIG_FILE);
 		    gdm_text_message_dialog (s);
+		    GdmPidFile = NULL;
 		    gdm_fail (_("%s: XDMCP disabled and no local servers defined. Aborting!"), "gdm_config_parse");
 	    }
     }
@@ -520,6 +527,7 @@ gdm_config_parse (void)
 		       "and restart gdm."),
 		     GDM_CONFIG_FILE);
 	    gdm_text_message_dialog (s);
+	    GdmPidFile = NULL;
 	    gdm_fail (_("%s: Can't find the gdm user (%s). Aborting!"), "gdm_config_parse", GdmUser);
     } else {
 	    GdmUserId = pwent->pw_uid;
@@ -533,6 +541,7 @@ gdm_config_parse (void)
 		       "correct gdm configuration %s and "
 		       "restart gdm."), GDM_CONFIG_FILE);
 	    gdm_text_message_dialog (s);
+	    GdmPidFile = NULL;
 	    gdm_fail (_("%s: The gdm user should not be root. Aborting!"), "gdm_config_parse");
     }
 
@@ -552,6 +561,7 @@ gdm_config_parse (void)
 		       "and restart gdm."),
 		     GDM_CONFIG_FILE);
 	    gdm_text_message_dialog (s);
+	    GdmPidFile = NULL;
 	    gdm_fail (_("%s: Can't find the gdm group (%s). Aborting!"), "gdm_config_parse", GdmGroup);
     } else  {
 	    GdmGroupId = grent->gr_gid;   
@@ -565,6 +575,7 @@ gdm_config_parse (void)
 		       "correct gdm configuration %s and "
 		       "restart gdm."), GDM_CONFIG_FILE);
 	    gdm_text_message_dialog (s);
+	    GdmPidFile = NULL;
 	    gdm_fail (_("%s: The gdm group should not be root. Aborting!"), "gdm_config_parse");
     }
 
@@ -609,6 +620,7 @@ gdm_config_parse (void)
 		       "restart gdm."), GdmServAuthDir,
 		     GDM_CONFIG_FILE);
 	    gdm_text_message_dialog (s);
+	    GdmPidFile = NULL;
 	    gdm_fail (_("%s: Authdir %s does not exist. Aborting."), "gdm_config_parse", GdmServAuthDir);
     }
 
@@ -621,6 +633,7 @@ gdm_config_parse (void)
 		       "restart gdm."), GdmServAuthDir,
 		     GDM_CONFIG_FILE);
 	    gdm_text_message_dialog (s);
+	    GdmPidFile = NULL;
 	    gdm_fail (_("%s: Authdir %s is not a directory. Aborting."), "gdm_config_parse", GdmServAuthDir);
     }
 
@@ -635,6 +648,7 @@ gdm_config_parse (void)
 		     GdmServAuthDir, GdmUser, GdmGroup,
 		     GDM_CONFIG_FILE);
 	    gdm_text_message_dialog (s);
+	    GdmPidFile = NULL;
 	    gdm_fail (_("%s: Authdir %s is not owned by user %s, group %s. Aborting."), "gdm_config_parse", 
 		      GdmServAuthDir, GdmUser, GdmGroup);
     }
@@ -650,6 +664,7 @@ gdm_config_parse (void)
 		       "restart gdm."),
 		     GdmServAuthDir, GDM_CONFIG_FILE);
 	    gdm_text_message_dialog (s);
+	    GdmPidFile = NULL;
 	    gdm_fail (_("%s: Authdir %s has wrong permissions %o. Should be 0750. Aborting."), "gdm_config_parse", 
 		      GdmServAuthDir, statbuf.st_mode);
     }
@@ -842,12 +857,10 @@ deal_with_x_crashes (GdmDisplay *d)
 	    if (pid == 0) {
 		    char *argv[2];
 		    char *xlog = g_strconcat (GdmLogDir, "/", d->name, ".log", NULL);
-		    int ii;
 
 		    closelog ();
 
-		    for (ii = 0; ii < sysconf (_SC_OPEN_MAX); ii++)
-			    close (ii);
+		    gdm_close_all_descriptors (0 /* from */, -1 /* except */);
 
 		    /* No error checking here - if it's messed the best response
 		    * is to ignore & try to continue */
@@ -991,22 +1004,22 @@ gdm_cleanup_children (void)
 
 	    if (d->sesspid > 0) {
 		    if (kill (d->sesspid, SIGTERM) == 0)
-			    waitpid (d->sesspid, NULL, 0);
+			    ve_waitpid_no_signal (d->sesspid, NULL, 0);
 		    d->sesspid = 0;
 	    }
 	    if (d->greetpid > 0) {
 		    if (kill (d->greetpid, SIGTERM) == 0)
-			    waitpid (d->greetpid, NULL, 0);
+			    ve_waitpid_no_signal (d->greetpid, NULL, 0);
 		    d->greetpid = 0;
 	    }
 	    if (d->chooserpid > 0) {
 		    if (kill (d->chooserpid, SIGTERM) == 0)
-			    waitpid (d->chooserpid, NULL, 0);
+			    ve_waitpid_no_signal (d->chooserpid, NULL, 0);
 		    d->chooserpid = 0;
 	    }
 	    if (d->servpid > 0) {
 		    if (kill (d->servpid, SIGTERM) == 0)
-			    waitpid (d->servpid, NULL, 0);
+			    ve_waitpid_no_signal (d->servpid, NULL, 0);
 		    d->servpid = 0;
 	    }
     }
@@ -1065,6 +1078,8 @@ gdm_cleanup_children (void)
 	    break;
     }
 
+start_autopsy:
+
     /* Autopsy */
     switch (status) {
 	
@@ -1084,33 +1099,45 @@ gdm_cleanup_children (void)
 	gdm_info (_("gdm_child_action: Master rebooting..."));
 
 	final_cleanup ();
+	chdir ("/");
 
 	argv = ve_split (GdmReboot);
 	execv (argv[0], argv);
 
 	gdm_error (_("gdm_child_action: Reboot failed: %s"), strerror (errno));
+
+	status = DISPLAY_REMANAGE;
+	goto start_autopsy;
 	break;
 	
     case DISPLAY_HALT:		/* Halt machine */
 	gdm_info (_("gdm_child_action: Master halting..."));
 
 	final_cleanup ();
+	chdir ("/");
 
 	argv = ve_split (GdmHalt);
 	execv (argv[0], argv);
 
 	gdm_error (_("gdm_child_action: Halt failed: %s"), strerror (errno));
+
+	status = DISPLAY_REMANAGE;
+	goto start_autopsy;
 	break;
 
     case DISPLAY_SUSPEND:	/* Suspend machine */
 	gdm_info (_("gdm_child_action: Master suspending..."));
 
 	final_cleanup ();
+	chdir ("/");
 
 	argv = ve_split (GdmSuspend);
 	execv (argv[0], argv);
 
 	gdm_error (_("gdm_child_action: Suspend failed: %s"), strerror (errno));
+
+	status = DISPLAY_REMANAGE;
+	goto start_autopsy;
 	break;
 
     case DISPLAY_RESTARTGDM:
@@ -1267,6 +1294,7 @@ static void
 signal_notify (int sig)
 {
 	gdm_signal_notify (sig);
+	g_main_context_wakeup (NULL);
 }
 
 /*
@@ -1320,6 +1348,56 @@ struct poptOption options [] = {
 	{ NULL, 0, 0, NULL, 0}
 };
 
+static gboolean
+linux_only_is_running (pid_t pid)
+{
+	char resolved_self[PATH_MAX];
+	char resolved_running[PATH_MAX];
+
+	char *running = g_strdup_printf ("/proc/%lu/exe", (gulong)pid);
+
+	if (realpath ("/proc/self/exe", resolved_self) == NULL) {
+		g_free (running);
+		/* probably not a linux system */
+		return FALSE;
+	}
+
+	if (realpath (running, resolved_running) == NULL) {
+		g_free (running);
+		/* probably not a linux system */
+		return FALSE;
+	}
+
+	g_free (running);
+
+	if (strcmp (resolved_running, resolved_self) == 0)
+		return TRUE;
+	return FALSE;
+}
+
+static void
+ensure_desc_012 (void)
+{
+	int fd;
+	/* We here ensure descriptors 0, 1 and 2
+	 * we of course count on the fact that open
+	 * opens the lowest available descriptor */
+	for (;;) {
+		fd = open("/dev/null", O_RDWR);
+		/* what to do on fail?  I dunno,
+		 * just keep on going is the safest
+		 * bet I suppose */
+		if (fd < 0)
+			break;
+		/* Once we are up to 3, we're beyond stdin,
+		 * stdout and stderr */
+		if (fd >= 3) {
+			close (fd);
+			break;
+		}
+	}
+}
+
 int 
 main (int argc, char *argv[])
 {
@@ -1329,6 +1407,9 @@ main (int argc, char *argv[])
     poptContext ctx;
     int nextopt;
 
+    /* We here ensure descriptors 0, 1 and 2 */
+    ensure_desc_012 ();
+
     store_argv (argc, argv);
 
     bindtextdomain (GETTEXT_PACKAGE, GNOMELOCALEDIR);
@@ -1377,7 +1458,8 @@ main (int argc, char *argv[])
 
         if (pf != NULL &&
 	    fscanf (pf, "%d", &pidv) == 1 &&
-	    kill (pidv, 0) != -1) {
+	    kill (pidv, 0) != -1 &&
+	    linux_only_is_running (pidv)) {
 		/* make sure the pid file doesn't get wiped */
 		GdmPidFile = NULL;
 		gdm_fail (_("gdm already running. Aborting!"));
@@ -1568,6 +1650,77 @@ gdm_signal_notify (gint8 signal)
 	signals_notified[index] |= 1 << shift;
 }
 
+static gboolean
+order_exists (int order)
+{
+	GSList *li;
+	for (li = displays; li != NULL; li = li->next) {
+		GdmDisplay *d = li->data;
+		if (d->x_servers_order == order)
+			return TRUE;
+	}
+	return FALSE;
+}
+
+static int
+get_new_order (GdmDisplay *d)
+{
+	int order;
+	GSList *li;
+	/* first try the position in the 'displays' list as
+	 * our order */
+	for (order = 0, li = displays; li != NULL; order++, li = li->next) {
+		if (li->data == d)
+			break;
+	}
+	/* next make sure it's unique */
+	while (order_exists (order))
+		order++;
+	return order;
+}
+
+static void
+write_x_servers (GdmDisplay *d)
+{
+	FILE *fp;
+	char *file = g_strconcat (GdmServAuthDir, "/", d->name, ".Xservers", NULL);
+	int i;
+	int bogusname;
+
+	if (d->x_servers_order < 0)
+		d->x_servers_order = get_new_order (d);
+
+	fp = fopen (file, "w");
+	if (fp == NULL) {
+		gdm_error ("Can't open %s for writing", file);
+		g_free (file);
+		return;
+	}
+
+	for (bogusname = 0, i = 0; i < d->x_servers_order; bogusname++, i++) {
+		char buf[32];
+		g_snprintf (buf, sizeof (buf), ":%d", bogusname);
+		if (strcmp (buf, d->name) == 0)
+			g_snprintf (buf, sizeof (buf), ":%d", ++bogusname);
+		fprintf (fp, "%s local /usr/X11R6/bin/Xbogus\n", buf);
+	}
+
+	if (d->type == TYPE_XDMCP) {
+		fprintf (fp, "%s foreign\n", d->name);
+	} else {
+		char **argv;
+		char *command;
+		argv = gdm_server_resolve_command_line
+			(d, FALSE /* resolve_handled */);
+		command = g_strjoinv (" ", argv);
+		g_strfreev (argv);
+		fprintf (fp, "%s local %s\n", d->name, command);
+		g_free (command);
+	}
+
+	fclose (fp);
+}
+
 static void
 send_slave_ack (GdmDisplay *d)
 {
@@ -1900,6 +2053,24 @@ gdm_handle_message (GdmConnection *conn, const char *msg, gpointer data)
 			if (d->greetpid > 0)
 				kill (d->greetpid, SIGHUP);
 		}
+	} else if (strncmp (msg, GDM_SOP_WRITE_X_SERVERS " ",
+		            strlen (GDM_SOP_WRITE_X_SERVERS " ")) == 0) {
+		GdmDisplay *d;
+		long slave_pid;
+
+		if (sscanf (msg, GDM_SOP_WRITE_X_SERVERS " %ld",
+			    &slave_pid) != 1)
+			return;
+
+		/* Find out who this slave belongs to */
+		d = gdm_display_lookup (slave_pid);
+
+		if (d != NULL) {
+			write_x_servers (d);
+
+			/* send ack */
+			send_slave_ack (d);
+		}
 	}
 }
 
@@ -2100,6 +2271,7 @@ check_cookie (const char *file, const char *disp, const char *cookie)
 
 static void
 handle_flexi_server (GdmConnection *conn, int type, const char *server,
+		     gboolean handled,
 		     const char *xnest_disp, uid_t xnest_uid,
 		     const char *xnest_auth_file,
 		     const char *xnest_cookie)
@@ -2172,6 +2344,7 @@ handle_flexi_server (GdmConnection *conn, int type, const char *server,
 				      "ERROR 2 Startup errors\n");
 		return;
 	}
+	display->handled = handled;
 
 	if (type == TYPE_FLEXI_XNEST) {
 		GdmDisplay *parent;
@@ -2506,6 +2679,7 @@ gdm_handle_user_message (GdmConnection *conn, const char *msg, gpointer data)
 			return;
 		}
 		handle_flexi_server (conn, TYPE_FLEXI, GdmStandardXServer,
+				     TRUE /* handled */,
 				     NULL, 0, NULL, NULL);
 	} else if (strncmp (msg, GDM_SUP_FLEXI_XSERVER " ",
 		            strlen (GDM_SUP_FLEXI_XSERVER " ")) == 0) {
@@ -2547,7 +2721,7 @@ gdm_handle_user_message (GdmConnection *conn, const char *msg, gpointer data)
 		}
 		g_free (name);
 
-		handle_flexi_server (conn, TYPE_FLEXI, command,
+		handle_flexi_server (conn, TYPE_FLEXI, command, svr->handled,
 				     NULL, 0, NULL, NULL);
 	} else if (strncmp (msg, GDM_SUP_FLEXI_XNEST " ",
 		            strlen (GDM_SUP_FLEXI_XNEST " ")) == 0) {
@@ -2576,6 +2750,7 @@ gdm_handle_user_message (GdmConnection *conn, const char *msg, gpointer data)
 		}
 		
 		handle_flexi_server (conn, TYPE_FLEXI_XNEST, GdmXnest,
+				     TRUE /* handled */,
 				     dispname, uid, xauthfile, cookie);
 
 		g_free (dispname);
@@ -2615,9 +2790,11 @@ gdm_handle_user_message (GdmConnection *conn, const char *msg, gpointer data)
 		gdm_connection_write (conn, "OK");
 		for (li = displays; li != NULL; li = li->next) {
 			GdmDisplay *disp = li->data;
-			gdm_connection_printf (conn, "%s%ld",
-					       sep, (long)disp->greetpid);
-			sep = ";";
+			if (disp->greetpid > 0) {
+				gdm_connection_printf (conn, "%s%ld",
+						       sep, (long)disp->greetpid);
+				sep = ";";
+			}
 		}
 		gdm_connection_write (conn, "\n");
 	} else if (strncmp (msg, GDM_SUP_UPDATE_CONFIG " ",
