@@ -29,6 +29,8 @@ void gdm_error  (const gchar *format, ...) G_GNUC_PRINTF (1, 2);
 void gdm_debug  (const gchar *format, ...) G_GNUC_PRINTF (1, 2);
 
 void gdm_fdprintf  (int fd, const gchar *format, ...) G_GNUC_PRINTF (2, 3);
+int gdm_fdgetc     (int fd);
+char *gdm_fdgets   (int fd);
 
 /* Note that these can actually clear environment without killing
  * the LD_* env vars if --preserve-ld-vars was passed to the
@@ -71,6 +73,8 @@ void gdm_desetuid (void);
 
 gboolean gdm_test_opt (const char *cmd, const char *help, const char *option);
 
+void gdm_close_all_descriptors (int from, int except);
+
 #endif /* GDM_MISC_H */
 
 /* EOF */
